#	CERG
#	Bilal Habib
#	March 15, 2016
#   This file contains all the five schemes used for PUF response generation.
# 	It also contains the three functions to calculate uniqueness, uniformity and reliability.

import os
import errno
import csv
import pstats
import sys, getopt
import itertools
from itertools import combinations
import math
import cProfile
from copy import deepcopy
import config
import glob



def	file_read(inputfile, raw_data):				#Read the inputfile and store in an array named raw_data
	with open(inputfile, "r") as f:
		reader = csv.reader(f, delimiter=",")
		
		for line in reader:
			raw_data.append( (list(line)))	
	return  raw_data
	
def components_read(chips, chip_num, ro_freq, FPGA_num):				
	print ('Chip identified: ' + repr(chips[chip_num][0]))
	FPGA_num.append(chips[chip_num][0])
	del chips[chip_num][0]
	
	for j in range(0,len(chips[chip_num])):
		result = (chips[chip_num][j])
		ro_freq.append(result)
	return ro_freq




def main():
	inputfile = config.inputfile						# input file for raw data
	num_devices = config.num_devices
	Response_bit_length = config.Response_bit_length
	Num_components_PC  = config.Num_components_PC
	Num_components_CNC = config.Num_components_CNC
	Num_components_BLG = config.Num_components_BLG
	Num_components_S_ArbRO = config.Num_components_S_ArbRO
	Num_components_Id_Map = config.Num_components_Id_Map
	set_size = config.set_size							#Parameter used in BLG scheme
	S_ArbRO_K = config.S_ArbRO_K						#Parameter used in S-ArbRO-2 scheme
	Id_Map_input_file = config.Id_Map_input_file
	Id_Map_q = config.Id_Map_q	 
	Id_Map_e = config.Id_Map_e	
	Id_Map_t = config.Id_Map_t	
	field_directory = config.field_directory
	field_nominal = config.field_nominal
	field_nominal_file = config.field_nominal_file
	
	
	################################ Create New Directory to store the result of schemes #######################################
		
	raw_data = []							# Store the inputdata
	file_read(inputfile, raw_data)
	response =[[] for i in range(num_devices)]
	
	input_data=str.find(inputfile,'_')
	input_data= inputfile[:input_data]
	
	
	current_loc = os.getcwd()
	newdir = input_data+'_Results_1'
	
	all_subdirs = [d for d in os.listdir('.') if os.path.isdir(d)]
	latest = max(all_subdirs, key=os.path.getctime)
		
	latest_data=str.find(latest,'_')
	latest_data= latest[:latest_data]
	
	
	
	if(latest_data== input_data):
		if not os.path.exists(newdir):
			os.makedirs(newdir)
		else:
			newdir = max(all_subdirs, key=os.path.getctime)
			num = int(''.join(ele for ele in newdir if ele.isdigit()))
			newdir = ''.join([i for i in newdir if not i.isdigit()])
	
			num=num+1
			newdir = newdir+str(num)
			os.makedirs(newdir)
	else:
		if not os.path.exists(newdir):
			os.makedirs(newdir)
		else:
			newdir = max( glob.glob(input_data+"*"), key=os.path.getctime)
			num = int(''.join(ele for ele in newdir if ele.isdigit()))
			newdir = ''.join([i for i in newdir if not i.isdigit()])
	
			num=num+1
			newdir = newdir+str(num)
			os.makedirs(newdir)
	
	print('All results stored at ' + current_loc +'\\'+newdir)
	

	################################ Call the schemes #######################################

	CNC( deepcopy(raw_data) , input_data+'_CNC.txt' , num_devices , Num_components_CNC, response, newdir) 		#inputdata, outputfile, No. of devices, components, response, directory_name
	uniqueness(response, input_data+'_uniqueness_CNC.txt', num_devices, Response_bit_length, newdir)		 					#inputdata, outputfile, No. of devices, response_bits, directory_name
	uniformity(response, input_data+'_uniformity_CNC.txt', num_devices, Response_bit_length, newdir)		 					#inputdata, outputfile, No. of devices, response_bits, directory_name
	entropy(response, input_data+'_entropy_CNC.txt', num_devices, Response_bit_length, newdir)		 							#inputdata, outputfile, No. of devices, response_bits, directory_name
	
	response =[[] for i in range(num_devices)]
	PC(deepcopy(raw_data) , input_data+'_PC.txt' , num_devices , Num_components_PC, response, newdir)   	#inputdata, outputfile, No. of devices, components, response, Directory_name
	uniqueness(response, input_data+'_uniqueness_PC.txt', num_devices, Response_bit_length, newdir)		 					#inputdata, outputfile, No. of devices, response_bits, directory_name
	uniformity(response, input_data+'_uniformity_PC.txt', num_devices, Response_bit_length, newdir)		 					#inputdata, outputfile, No. of devices, response_bits, directory_name
	entropy(response, input_data+'_entropy_PC.txt', num_devices, Response_bit_length, newdir)		 							#inputdata, outputfile, No. of devices, response_bits, directory_name
		
	response =[[] for i in range(num_devices)]
	BLG(deepcopy(raw_data), input_data+'_BLG.txt',  num_devices, Num_components_BLG, set_size, response, newdir)	#inputdata, outputfile, No. of devices, components, set_size, response, directory_name
	uniqueness(response, input_data+'_uniqueness_BLG.txt', num_devices, Response_bit_length, newdir)		 					#inputdata, outputfile, No. of devices, response_bits, directory_name
	uniformity(response, input_data+'_uniformity_BLG.txt', num_devices, Response_bit_length, newdir)		 					#inputdata, outputfile, No. of devices, response_bits, directory_name
	entropy(response, input_data+'_entropy_BLG.txt', num_devices, Response_bit_length, newdir)		 							#inputdata, outputfile, No. of devices, response_bits, directory_name
	
	response =[[] for i in range(num_devices)]
	S_ArbRO_2(deepcopy(raw_data), input_data+'_SArbRO.txt', num_devices, Num_components_S_ArbRO, S_ArbRO_K, response, newdir)	#inputdata, outputfile, No. of devices, components, K, response, directory_name
	uniqueness(response, input_data+'_uniqueness_SArbRO.txt', num_devices, Response_bit_length, newdir)		 					#inputdata, outputfile, No. of devices, response_bits, directory_name
	uniformity(response, input_data+'_uniformity_SArbRO.txt', num_devices, Response_bit_length, newdir)		 					#inputdata, outputfile, No. of devices, response_bits, directory_name
	entropy(response, input_data+'_entropy_SArbRO.txt', num_devices, Response_bit_length, newdir)		 							#inputdata, outputfile, No. of devices, response_bits, directory_name
	
	response =[[] for i in range(num_devices)]
	identity_map_all(deepcopy(raw_data), input_data+'_Enrol_R_Id_Map_all.txt', num_devices, Num_components_Id_Map, Id_Map_q, Id_Map_t, Id_Map_e, response, newdir)# identity_map_all.py -i 				1.2V_25C.csv -o Enrol_R.txt 	30 	2 	0.5 	0			17
	uniqueness(response, input_data+'_uniqueness_Id_Map.txt', num_devices, Response_bit_length, newdir)
	uniformity(response, input_data+'_uniformity_Id_Map.txt', num_devices, Response_bit_length, newdir)		 					#inputdata, outputfile, No. of devices, response_bits
	entropy(response, input_data+'_entropy_Id_Map.txt', num_devices, Response_bit_length, newdir)		 							#inputdata, outputfile, No. of devices, response_bits, directory_name
	
	response =[[] for i in range(num_devices)]
	#		To calculuate Reliability using Identity Mapping, un-comment the following line
	#rel_id_map(newdir, current_loc, field_directory, field_nominal, Id_Map_input_file, Num_components_Id_Map, Id_Map_q, Id_Map_t, Id_Map_e, input_data, Response_bit_length)
	#return
	#################################### Reliability at Field  ###############################
	newdir_field = 'Field_results'			#Store Field Responses in this Director.
	os.makedirs(os.path.join(newdir, newdir_field))
	newdir_field = newdir + '/' + newdir_field
		
	Conditions_field = 0
	path = os.path.join(current_loc, field_directory)
	print('path =' + path)
	
	new_file = []	# Stores the file names in the field
	if not os.path.exists(field_directory):
		print('Field_directory not found\n')
	
	for information in os.walk(path):
		Conditions_field += len(information[2])	# Count the number of files in the field.
	
	for i in range(0,Conditions_field):
		if  (str(information[2][i]) != field_nominal):
			x=str.find(str(information[2][i]), '.csv')
			new_file.append( str(information[2][i])[:x])
		
			
	print('Total conditions at field ' + repr(Conditions_field-1))
	
	
	raw_data = []							# Store the data of field device.
	file_path = os.path.join(field_directory, field_nominal)
	file_read(file_path, raw_data)
		
	trim=str.find(field_nominal,'.csv')
	field_nominal= field_nominal[:trim]
	
	total_devices_field = len(raw_data)
	Field_response =[[] for i in range(total_devices_field)]
	response =[[] for i in range(total_devices_field)]
	
	print('total_devices ' + repr(total_devices_field))
	#Un-comment a single scheme for generating result at Field Nominal Conditions.
	CNC( deepcopy(raw_data) , field_nominal+'_Field_CNC.txt' , total_devices_field , Num_components_CNC, response, newdir_field) 
	#PC( deepcopy(raw_data) , field_nominal+'_Field_PC.txt' , total_devices_field , Num_components_PC, response, newdir_field) 
	#S_ArbRO_2(deepcopy(raw_data), field_nominal+'_SArbRO.txt', total_devices_field, Num_components_S_ArbRO, S_ArbRO_K, response, newdir_field)
	#BLG(deepcopy(raw_data), field_nominal+'_BLG.txt',  total_devices_field, Num_components_BLG, set_size, response, newdir_field)
		
	raw_data_f1 = [[] for i in range(Conditions_field-1)]	#Read all the files in the field, except the nominal one (e.g 1.2V_25C)
	for i in range(len(new_file)):	
		file_path = os.path.join(field_directory, (new_file[i]+'.csv'))
		file_read(file_path, raw_data_f1[i])
	print(len(raw_data_f1))
	
	
	
	for i in range(0,Conditions_field-1):
		#Un-comment a single scheme for generating result at all the Field Conditions.
		#S_ArbRO_2(deepcopy(raw_data_f1[i]), new_file[i]+'_SArbRO.txt', total_devices_field, Num_components_S_ArbRO, S_ArbRO_K, Field_response, newdir_field)
		#BLG(deepcopy(raw_data_f1[i]), new_file[i]+'_BLG.txt',  total_devices_field, Num_components_BLG, set_size, Field_response, newdir_field)
		#PC( deepcopy(raw_data_f1[i]) , new_file[i]+'_Field_PC.txt' , total_devices_field , Num_components_PC, Field_response, newdir_field)
		CNC( deepcopy(raw_data_f1[i]) , new_file[i]+'_Field_CNC.txt' , total_devices_field , Num_components_CNC, Field_response, newdir_field) 
		reliability(response, Field_response, new_file[i]+'_Reliability_.txt', total_devices_field, Response_bit_length, newdir_field) #inputdata_nominal, inputdata_field, output_file, Num_devices_field, Response_lenght, directory_name
		Field_response =[[] for i in range(total_devices_field)]
	
	return
	
	
###################CNC scheme starts###########################

def id_gen_pc(ro_freq, ID_size, response):	
	for i in range(0, ID_size):
		if (int(float(ro_freq[i]))>int(float(ro_freq[i+1]))):
			response.append(1)
		else:
			response.append(0)
	
	return response
	

 
def CNC(chips, outputfile, devices, components, response, newdir):    
	print ('Script used for Pairwise comparison')
	print ('Output file is ' + repr(outputfile))

	N = devices
	
	m = components - 1 # N-1 bits generated from N components
	ro_freq=[]
	chip_num = 0
	print('Number of devices ' + repr(N))
	print ('Components per chip: ' + repr(components) + '\n')

	ID_size = m
	FPGA_num=[]
	x = 0
	fi = open(os.path.join(newdir, outputfile), 'w')
	
	while x<N  :
		ro_freq = []
		components_read(chips, x, ro_freq, FPGA_num)
		id_gen_pc(ro_freq, ID_size, response[x])
		print('Length of response ' + repr(len(response[x])))
		print ('Response generated for chip#:' + repr(FPGA_num[x]) + '\n')	#,"size of resp", len(response) 
		str_out = ''.join([str(y) for y in response[x]])
		fi.write(str_out) 			
		fi.write('\n')
		x= x + 1
				
	print ('PUF IDs written to file: ' + (fi.name)	)
	print ('-------------------------------------------------\n')
	fi.close()
	return 
###################  End CNC  ###########################

################### PC Scheme starts ###########################


def id_gen_eco(ro_freq, ID_size, response):
	i = 0
	
	while (i< (2 * ID_size)):
		if (int(float(ro_freq[i])) > int(float(ro_freq[i + 1]))):
			response.append(1)
		else:
			response.append(0)
		
		i = i+2	
	
	return response

   
def PC(chips, outputfile, devices, components, response, newdir):    
	print ('Script used for Pairwise comparison, Each Component compared Once')
	print ('Output file is ' + repr(outputfile))

	N= devices

	m= components
	
	print('Number of chips ' + repr(N))
	print ('Components per chip: ' + repr(m))
	
	fpgas = N	
	print ('Total Components per chip:' + repr( m))	
	ID_size = math.floor(m/2)
	print ('ID_size is ' + repr(ID_size) + '\n')
	x=0		#Select column in Sheet1 
	
	FPGA_num=[]
	
	fi = open(os.path.join(newdir, outputfile), 'w')
	
	while x < fpgas :
		ro_freq = []
		
		components_read(chips, x, ro_freq, FPGA_num)
				
		id_gen_eco(ro_freq, ID_size, response[x])
		print('Length of response ' + repr(len(response[x])))
		print ('Response generated for chip#: ' + repr(FPGA_num[x]) + '\n')	#,"size of resp", len(response) 
		str_out = ''.join([str(x) for x in response[x]])
		fi.write(str_out) 			
		fi.write('\n')		
		x= x + 1
		
	print ('PUF IDs written to file: ' + (fi.name)	)
	print ('-------------------------------------------------\n')
	fi.close()
	return
################### PC ends ###########################

################### Binary Lehmer_Gray  starts ###########################

def Gray(response):							#To do the Gray Encoding
	olist = [''.join(each) for each in zip(*[iter(response)]*1)]	#convert to string array
	
	temp = []
	temp.append(int(olist[0]))
	for p in range (1,len(olist)):
		x = int(olist[p]) ^ int(olist[p-1])
		temp.append(x)
	response = temp
	#print(response)
	return response
	
def Lehmer_sorting(sets, response):			#To do the Lehmer Encoding
	temp = sets
	y=0
	#print(temp)
	
	for j in range(1,len(temp)):
		z = 0
		for i in range(0,j):
			if (int(float(temp[j]))< int(float(temp[i]))):
				z = z +1
		if((j > 1) and (j <= 3) and (z < 2)):
			p = format(z, '#004b')	#contribute 2 bits to response
			response.extend(Gray(str(p[2:])))
		elif((j > 3) and (j<=7) and (z < 4)):
			p = format(z, '#005b')	#contribute 3 bits to response
			response.extend(Gray(str(p[2:])))
		elif((j > 7) and (j<= 15) and (z < 8)):
			p = format(z, '#006b')	#contribute 4 bits to response
			response.extend(Gray(str(p[2:])))
			
		elif((j > 15) and (j<=31) and (z < 16)):
			p = format(z, '#007b')	#contribute 5 bits to response
			response.extend(Gray(str(p[2:])))
		else:

			p = bin(z)
			response.extend(Gray(str(p[2:])))
		#print(response)
	#sys.exit()
	for i in range (0 , len(response)):
		response[i]= str(response[i])
	return response


def id_gen_lg(ro_freq , response, m, set_size):	
	sets = []
	Extended_R = []
	x= 0
	y= set_size
	temp=list(ro_freq)
	for k in range (0,int(m/set_size)):
		for i in range (x,y):
			sets=list(temp[x:y])
		x = x + set_size
		y = y + set_size
		Extended_R = Lehmer_sorting(sets, response)
	return Extended_R

def BLG(chips, outputfile,  devices, components , set_size, response, newdir ):
	N = devices
	m = components   # total components
	
	print ('Script used for Lehmer-Gray Encoding')
	print ('Output file is ' + repr(outputfile))
	print('Set size ' + repr(set_size))
	print('Total Devices ' + repr(N))
	print ('Components per chip: ' + repr(m) + '\n')
	
	ro_freq=[]
	chip_num = 0
	x=0		#Select column in Sheet1 
	
	FPGA_num=[]
	fi = open(os.path.join(newdir, outputfile), 'w')
		
	while x<N:
		ro_freq=[]
		components_read(chips, x, ro_freq, FPGA_num)
		
		id_gen_lg(ro_freq, response[x], m, set_size)
		print ('ID size ' + repr(len(response[x])))
		str_out = ''.join([str(x) for x in response[x]])
		fi.write(str_out) 			
		fi.write('\n')
		x= x + 1
	
		
	print ('\nPUF IDs written to file: ' + (fi.name)	)
	print ('-------------------------------------------------\n')
	fi.close()	
	return
################### Lehmer_Gray Ends ###########################

################### S-ArbRO-2 starts ##################################
def sub (a,b):
	return float(a)-float(b)
	
def id_gen_s(E, K,Kp, response):
	index=[]
	sums=[]
	for x in itertools.combinations(E, K):	#Generate all combinations of choose K.
		index.append(list(x))
	for G1 in range(len(index)):	# Generate Challenges for G1
		for G2 in range(Kp):		# Generate Challenges for G2
			p=list('0'*(K-len(bin(G2)[2:]))+bin(G2)[2:])	#remove 0b from any binary number.
			temp=[]
			
			for s in range(len(p)):
				if(p[s] =='0'):
					temp.append(sub(index[G1][s][0],index[G1][s][1]))	# r1-r2
				else:
					temp.append(sub(index[G1][s][1],index[G1][s][0])) 	# r2-r1
			#print('temp:' + repr(temp))
			sums.append(sum(temp))
	#print('sums: ' + repr(sums))	
	for i in range(0,len(sums)):
		if(sums[i]>=0):
			response.append(1)
		else:
			response.append(0)
	return response
	
  
def S_ArbRO_2(chips, outputfile, devices, components, K, response, newdir):
	print('Script used for S_ArbRO_2 ')
	N = devices
	m = components
	print(outputfile)
	print(newdir)
	print(chips[0][1])
	#sys.exit()
	
	ro_freq=[]
	chip_num = 0
	print('Number of chips ' + repr(N))
	print ('Components per chip: ' + repr(m))

	print ('Elements: ' + repr( m/2))
	
	
	
	print ('K chosen: ' + repr(K) + '\n')	
	Kp= pow(2,(K-1))
	x=0		
	
	fi = open(os.path.join(newdir, outputfile), 'w')
	FPGA_num=[]
	
	while x < N:
		ro_freq = []
		components_read(chips, x, ro_freq, FPGA_num)
		E= [[0 for i in range(2)] for j in range(int(m/2))]
	
		i=0
		while(i< m-1):				#Conversion to Elements
			E[int(i/2)][0]=ro_freq[i]
			E[int(i/2)][1]=ro_freq[i+1] 
			i=i+2
	
		
		id_gen_s(E, K, Kp, response[x])			#Generate ID
		ID_size = len(response[x])
		print ('ID_size is ' + repr(ID_size) + '\n')
		str_out = ''.join([str(x) for x in response[x]])
		fi.write(str_out) 			
		fi.write('\n')
		
		x= x + 1
	print ('\nPUF IDs written to file: ' + (fi.name)	)
	print ('-------------------------------------------------\n')
	fi.close()
	return
################### S-ArbRO-2 Ends ########################################

################### Id-Map  ########################################
def components_read_id_map(chips, chip_num, ro_freq, FPGA_num, m):
	print ('Chip identified: ' + repr(chips[chip_num][0]))
	FPGA_num.append(chips[chip_num][0])
	#print(chips[chip_num])
	del chips[chip_num][0]
	#print(chips[chip_num])
	
	for j in range(0,m):
		result = (chips[chip_num][j])
		ro_freq.append(int(float(result)))
	return ro_freq
	
def	file_read_field(fname, f_chips):	
	with open(fname, "r") as f:
		reader = csv.reader(f, delimiter=",")
		
		for line in reader:
			f_chips.append( (list(line)))	
	return  f_chips
	
def components_read_id_map_field(chips, chip_num, ro_freq, FPGA_num, m):
	print ('Chip identified: ' + repr(chips[0]))
	#FPGA_num.append(chips[chip_num][0])
	#print(chips)
	del chips[0]
	#print(chips)
	
	for j in range(0,m):
		result = (chips[j])
		ro_freq.append(int(float(result)))
	
	return ro_freq
	
def	file_read_enrol(inputfile, en_chips):	
	with open(inputfile, "r") as f:
		reader = csv.reader(f, delimiter=",")
		
		for line in reader:
			en_chips.append( (list(line)))	
	return  en_chips

def choose(n, r):
	return math.factorial(n)/(math.factorial(n-r)* math.factorial(r))
  
def f(x1, x2, e):
	temp = math.fabs(x1 - x2)
	return math.pow(temp,e)

	
def identity_map(inputfile, Id_Map_field_file,	 enrol_outputfile, field_outputfile, Field_response,  row, components, q, t, e, response, newdir, input_data):    
	print('------------------------------------------')
	print('Script used for Identity Mapping')
	en_chips=[]			#Data of all chips during enrolment
	ro_freq_enrol = []	#Data of a single chip
	file_read_enrol(inputfile, en_chips)
	
	print('Enrolment file is ' + repr(inputfile))
	
	N = len(en_chips)		# total FPGAs
	print('Total chips available: ' + repr(N))
	
	m = components
	print ('Components per chip: ' + repr(m) + '\n')
	FPGA_num=[]
	
	
	chip = row
	print ('Row selected during enrolment: ' + repr(chip))
	
	
	components_read_id_map(en_chips, chip, ro_freq_enrol, FPGA_num, m)
	#print(ro_freq_enrol)
	
	
	print ('Bucket size chosen: ' + repr(q))	
	
	print ('Parameter , t =  ' + repr(t) )
	
	
	print ('Parameter , e =  ' + repr(e) )	
	

	S = [x for r in range(t, t+1) for x in itertools.combinations(ro_freq_enrol, r)]
	Q = [sum([f(a, b, e) for a,b in itertools.combinations(x, 2)]) for x in S]
	R = [int(math.fmod(Q_i/q, 2)) for Q_i in Q]
	Wt = [0.5*q - (d - q*math.floor(float(d)/q)) for d in Q]
		
	#response[0].extend(R)
	
	response.extend(R)
	
	file_template = input_data+'_Enrol_{0}.txt'
	data = ['Q_Id_Map', 'Wt_Id_Map']
				
	for (fname, seq) in zip(data, [Q, Wt]):
		fi = open(os.path.join(newdir, file_template.format(fname)), 'w')
		str_out = '\n'.join([str(x) for x in seq])
		fi.write(str_out)
		fi.close()
			
	str_out = ''.join(str(e) for e in R)
	f2= open(os.path.join(newdir, enrol_outputfile),'w')
	f2.write(str_out)
	f2.close()
	
	print ('\nPUF bits generated: ' + repr(len(R)))
	print ('PUF IDs written to file :' + ( enrol_outputfile))
	print ('PUF Qs written to file : Enrol_Q')
	print ('Helper Data written to file : Enrol_Wt.txt')
	
	
		
	##else:		######################### Error Correction at the Field####################
		
		
	
	chip = row 		#	Row of field data
	print ('\nRow selected, field data: ' + repr(chip))
	
	f_chips=[]			#Data of all chips in the field
	ro_freq_field = []	#Data of a single chip
	
	
	
	r_corrected=[]
	temp=[]
	Q=[]	# Q will be generated again in the field
	FPGA_num=[]
		
	components_read_id_map_field(Id_Map_field_file, chip, ro_freq_field, FPGA_num, m)
	#print(ro_freq_field)
	

	#print(ro_freq_field)
	
	
	S = [x for r in range(t, t+1) for x in itertools.combinations(ro_freq_field, r)]
	Q = [sum([f(a, b, e) for a,b in itertools.combinations(x, 2)]) for x in S]
	
	for i in range (0,len(Q)):	#Q generated in the field
		temp.append( float(Wt[i]) + float(Q[i]))

	r_corrected = [int(math.fmod(Q_i/q, 2)) for Q_i in temp]
	Field_response.extend( r_corrected)
	
	fi = open(os.path.join(newdir, field_outputfile), 'w')	
	str_out = ''.join([str(x) for x in r_corrected])
	print ('\nField response written to file :' + field_outputfile)
	fi.write(str_out)
	print(fi.name)
	#sys.exit()
	print ('-------------------------------------------------\n')
	fi.close()
	return
################### Id-Map Ends ########################################

################## Id-Map for all devices starts #######################
#No error correction included
def components_read_all(chips, chip_num, ro_freq, FPGA_num,m):
	print ('chip identified: ' + repr(chips[chip_num][0]))
	FPGA_num.append(chips[chip_num][0])
	del chips[chip_num][0]
	for j in range(0,m):
		result = (chips[chip_num][j])
		ro_freq.append(int(float(result)))
	return ro_freq
		
	
def identity_map_all(chips, enrol_outputfile,devices,  components, q, t, e, response, newdir):      
	
	
	print('Script used for Identity Mapping for all devices at nominal conditions')
	en_chips=[]			#Data of all chips during enrolment
	ro_freq_enrol = []	#Data of a single chip
	
	
	N = devices		# total FPGAs
	print('Total chips available: ' + repr(N))
	
	m = components
	print ('Components per chip: ' + repr(m) + '\n')
	FPGA_num=[]
	
	print ('Bucket size chosen: ' + repr(q))	
	
	print ('Parameter , t =  ' + repr(t) )
	
	print ('Parameter , e =  ' + repr(e) )
	
	f2= open(os.path.join(newdir, enrol_outputfile),'w')
	for c in range (0,N):
		print(repr(c) + '\n')
		
		chip = int(c)
		R=[]
		Q=[]
		S=[]
		print ('Row selected: ' + repr(chip))
		ro_freq_enrol = []	#Data of a single chip
		components_read_all(chips, chip, ro_freq_enrol, FPGA_num,m)
		
		S = [x for r in range(t, t+1) for x in itertools.combinations(ro_freq_enrol, r)]
		Q = [sum([f(a, b, e) for a,b in itertools.combinations(x, 2)]) for x in S]
		R = [int(math.fmod(Q_i/q, 2)) for Q_i in Q]
		response[c].extend(R)		
		str_out = ''.join(str(e) for e in R)
		
		f2.write(str_out)
		f2.write('\n')
	f2.close()
	return
################## Id-Map for all devices ends  ########################

################### Uniqueness ############################################
def uniqueness(response, outputfile, devices, response_bits, newdir):
	print('\nUniqueness starts')	

	N = devices
	
	
	print('N = ' + repr(N) + ' devices')
		
	K = 1	 			# Number of Identifiers per FPGA
	L = response_bits 	# Number of bits per identifiers
	print ('L = ' +  repr(L) + ' bits')
	totalcnt = 0
	
	HD=[]					#Hamming Distance between any two chips
	totalcnt = 0
	for i in range (0,N-1):
		for p in range(i,N-1):
			diffs = 0
			for k in range(0,L):
				if(int(response[i][k]) != int(response[p+1][k])):
					diffs += 1
			HD.append(diffs)
			totalcnt+= diffs

	
	Uniqueness_Maiti = float(totalcnt)*2/(K*L*N*(N-1)) #uniqueness
	Uniqueness_Hori = float(totalcnt)*4/(K*L*N*N) #uniqueness
	min_HD = (min(HD)/L)*100
	max_HD = (max(HD)/L)*100
	WC_Uniqueness =float (min(min_HD, 100-max_HD))
	

	
	fi= open(os.path.join(newdir, outputfile), 'w')
	fi.write('The Uniqueness according to Maiti is ' + repr(float("{:.2f}".format(Uniqueness_Maiti * 100)))+'%')
	fi.write ('\nThe Uniqueness according to Hori is ' + repr (float("{:.2f}".format(Uniqueness_Hori * 100)))+'%')
	fi.write ('\nThe Worst Case Uniqueness is ' + repr (float("{:.2f}".format(WC_Uniqueness )))+'%')
	fi.write ('\nThe 50% - WC Uniq is ' + repr (float("{:.2f}".format(50.0 - WC_Uniqueness )))+'%')
	fi.write ('\nInter-chip min: ' + repr (float("{:.2f}".format(min_HD )))+'%')
	fi.write ('\nInter-chip max: ' + repr (float("{:.2f}".format(max_HD )))+'%')
	str_out = '\n'.join([str(x) for x in HD])
	fi.write('\nHamming Distances of : ' + repr(N) + ' chips is given,\n')
	fi.write(str_out)
	fi.close()
	print('Uniqueness results are written to ' + fi.name)
	print ('-------------------------------------------------\n')
	return
################### Uniqueness Ends ############################################

################### Uniformity ############################################
def uniformity(chips, outputfile, devices, response_bits, newdir):	
	
	print('Uniformity starts')	
			
	N = devices
	print('N =' + repr(N) + ' devices')
	
	fi= open(os.path.join(newdir, outputfile), 'w')
	ones_in_chip=[]
	sum = 0
	length= response_bits
	print('Bits per chip = ' + repr(length))
	for k in range(0, N):				
		ones_per_chip = 0
		for i in range (0, length):
			if(int(chips[k][i]) == 1):
				ones_per_chip += 1
		ones_in_chip.append(ones_per_chip)
		sum += ones_per_chip
		fi.write('Uniformity of chip No ' + repr(k)+ ' is ' + repr(float("{:.2f}".format(ones_per_chip/float(length)*100))) + '%\n')
	uniformity = sum/float(length*N) *100
		
	fi.write('Avg Uniformity for : '+ repr(N) + ' chips is \nUniformity: ' + repr(float("{:.2f}".format(uniformity))) + '%')
	str_out = '\n'.join([str(x) for x in ones_in_chip])
	fi.write('\nOnes in chip are: \n')
	fi.write(str_out)
	print('Uniformity results are written to ' + fi.name)
	print ('-------------------------------------------------\n')
	fi.close()
	return
################### Uniformity Ends ############################################	

################### Reliability starts #########################################
def reliability(response, Field_response, outputfile, devices, bits, newdir):
	print('Script for Reliability starts')
	#enrol = []
	#conditions = []
	
	#enrol=list(response)
	#conditions = list(Field_response)
		
	
	L = bits      			# Number of PUF bits generated per FPGA
	print('Response bits ' + repr(L))
		
	K = 1	#no of identifiers
	N = devices #len(conditions)	#no of operating conditions (NOT including the nominal one)
	
	
	T = 1	#no of samples taken
	fi= open(os.path.join(newdir, outputfile), 'w')	
	fi.write('Reliability result:  ' )
	total_rel=0
	totalcnt = 0
	rel_all=[]

	for i in range (0,N):
		diffs = 0
		for j in range (0,L-1):
			if(int(Field_response[i][j]) != int(response[i][j])):
				diffs += 1
		totalcnt+= diffs	
		rel_i = (1-float(diffs)/(K*T*L))*100
		fi.write ('\nThe reliability of device ' + repr(i)+ ' is ' + repr(float("{:.2f}".format(rel_i))) + '%')
		fi.write ('\n\tErroneous bits in device: '  + repr(diffs) + ' out of ' + repr(L) + ' bits')
		total_rel += rel_i
		rel_all.append(rel_i)
			
	avg_rel = float(total_rel/N) 	#reliability over N chips
	#avg_rel = float(total_rel) 	#reliability over 1 chips
	fi.write ('\n\nThe average Reliability of ' +  repr(N)   + ' devices : ' + repr(float("{:.2f}".format(avg_rel))) + '%')
	fi.write ('\n\nWorse Case Reliability : ' +  repr(float("{:.2f}".format(min(rel_all) ))) + '%')
	fi.write ('\n\n100% - WC Reliability : ' +  repr(float("{:.2f}".format(100.0 - min(rel_all) ))) + '%')
	print('\nReliability data stored in ' + (fi.name))
	fi.close
	return
################### Reliability ends   #########################################


################### Entropy starts   #########################################
def entropy(chips, outputfile, devices, response_bits, newdir):		
	
	print('Entropy starts')	
			
	N = devices
	print('N =' + repr(N) + ' devices')
	
	fi= open(os.path.join(newdir, outputfile), 'w')
	ones_in_chip=[]
	sum = 0
	length= response_bits
	print('Bits per chip = ' + repr(length))
		
	sum_bin_entropy = 0
	sum_joint_entropy = 0
	worse_bin_entropy = 1
	worse_joint_entropy = 1
	
	for k in range(0, N):
		sum = 0
		prob_1=0
		prob_0=0
		prob_00=0
		prob_01=0
		prob_10=0
		prob_11=0
		binary_entropy = 0
		joint_entropy = 0
		for i in range (0, length):
			ones_per_chip = 0
			if(int(chips[k][i]) == 1):
				ones_per_chip += 1
			sum += ones_per_chip
		i=0
		while(i < length-1):
			if(int(chips[k][i]) == 0 and int(chips[k][i+1])== 0):
				prob_00 += 1                                
			if(int(chips[k][i]) == 0 and int(chips[k][i+1])== 1):
				prob_01 += 1                                
			if(int(chips[k][i]) == 1 and int(chips[k][i+1])== 0):
				prob_10 += 1                                
			if(int(chips[k][i]) == 1 and int(chips[k][i+1])== 1):
				prob_11 += 1
			i = i + 2
		print(k)
		print ('prob_00 : '+ repr(prob_00) + ' prob_01 : '+ repr(prob_01)+ ' prob_10 : '+ repr(prob_10)+ ' prob_11 : '+ repr(prob_11) + '\n')
	
		prob_00 = float(2*prob_00/length)
		prob_01 = float(2*prob_01/length)
		prob_10 = float(2*prob_10/length)
		prob_11 = float(2*prob_11/length)
		prob_1 = float(sum/length)
		prob_0 = float(1-prob_1)
		binary_entropy = -1*prob_1*math.log(prob_1,2) - prob_0*math.log(prob_0,2)
		if(prob_00==0.0 and prob_01==0.0):
			joint_entropy = binary_entropy-(0 		+   0 	+   prob_10* math.log((prob_10/(prob_1*prob_0)),2) 	+   prob_11* math.log((prob_11/(prob_1*prob_1)),2))
		elif(prob_10==0.0 and prob_11==0.0):
			joint_entropy = binary_entropy-(prob_00* math.log((prob_00/(prob_0*prob_0)),2) 		+   prob_01* math.log((prob_01/(prob_0*prob_1)),2) 	+   0 												+   0)
		elif(prob_01==0.0 and prob_10==0.0):
			joint_entropy = binary_entropy-(prob_00* math.log((prob_00/(prob_0*prob_0)),2) 		+   0 												+   0 												+   prob_11* math.log((prob_11/(prob_1*prob_1)),2))
		elif(prob_00==0.0 and prob_11==0.0):
			joint_entropy = binary_entropy-(0 													+   prob_01* math.log((prob_01/(prob_0*prob_1)),2) 	+   prob_10* math.log((prob_10/(prob_1*prob_0)),2) 	+   0)
		elif(prob_01==0.0 and prob_11==0.0):
			joint_entropy = binary_entropy-(prob_00* math.log((prob_00/(prob_0*prob_0)),2) 		+   0 												+   prob_10* math.log((prob_10/(prob_1*prob_0)),2) 	+   0)
		elif(prob_00==0.0):
			joint_entropy = binary_entropy-(0 													+   prob_01* math.log((prob_01/(prob_0*prob_1)),2) 	+   prob_10* math.log((prob_10/(prob_1*prob_0)),2) +   prob_11* math.log((prob_11/(prob_1*prob_1)),2))
		elif(prob_01==0.0):
			joint_entropy = binary_entropy-(prob_00* math.log((prob_00/(prob_0*prob_0)),2) 		+   0 												+   prob_10* math.log((prob_10/(prob_1*prob_0)),2) 	+   prob_11* math.log((prob_11/(prob_1*prob_1)),2))
		elif(prob_10==0.0):
			joint_entropy = binary_entropy-(prob_00* math.log((prob_00/(prob_0*prob_0)),2) 		+   prob_01* math.log((prob_01/(prob_0*prob_1)),2) 	+   0 												+   prob_11* math.log((prob_11/(prob_1*prob_1)),2))
		elif(prob_11==0.0):
			joint_entropy = binary_entropy-(prob_00* math.log((prob_00/(prob_0*prob_0)),2) 		+   prob_01* math.log((prob_01/(prob_0*prob_1)),2) 	+   prob_10* math.log((prob_10/(prob_1*prob_0)),2) 	+   0)
		else:
			joint_entropy = binary_entropy-(prob_00* math.log((prob_00/(prob_0*prob_0)),2) 		+   prob_01* math.log((prob_01/(prob_0*prob_1)),2) 	+   prob_10* math.log((prob_10/(prob_1*prob_0)),2) 	+   prob_11* math.log((prob_11/(prob_1*prob_1)),2))
		
		print ('prob_1 : '+ repr(prob_1) )
		print ('prob_0 : '+ repr(prob_0) )
		sum_bin_entropy += binary_entropy
		sum_joint_entropy += joint_entropy
		if(binary_entropy < worse_bin_entropy):
			worse_bin_entropy = binary_entropy
		if(joint_entropy < worse_joint_entropy):
			worse_joint_entropy = joint_entropy
		#if(worse_joint_entropy<0.0):
		#	sys.exit()
	
	#print ('prob_00 : '+ repr(prob_00) + ' prob_01 : '+ repr(prob_01)+ ' prob_10 : '+ repr(prob_10)+ ' prob_11 : '+ repr(prob_11))

		
		print ('binary entropy : '+ repr(binary_entropy) )
		print ('joint entropy : '+ repr(joint_entropy) + '\n')
		
		fi.write('For chip # ' + repr(k) + ', Binary entropy : '+ repr(format(binary_entropy,'.10f')) + ' Joint entropy: ' + repr(format(joint_entropy,'.10f'))  + '\n')
	
	fi.write('Worse Joint entropy : '+ repr(format(worse_joint_entropy,'.10f'))+ '\n')
	fi.write('100% - WC Joint entropy : '+ repr(format(1 - worse_joint_entropy,'.10f'))+ '\n')
	fi.write('Avg Joint entropy : '+ repr(format(sum_joint_entropy/N,'.10f')) + ' For ' +repr(N) + ' devices,\n')
	fi.write('Worse Binary entropy : '+ repr(format(worse_bin_entropy,'.10f')) + '\n')
	fi.write('Avg Binary entropy : '+ repr(format(sum_bin_entropy/N,'.10f')) + ' For ' +repr(N) + ' devices'  )
	
	
	
		
	print('Entropy results are written to ' + fi.name)
	print ('-------------------------------------------------\n')
	fi.close()
	return
################### Entropy ends     #########################################
#################################### Reliability for ID-Map  ###############################
def rel_id_map(newdir, current_loc, field_directory, field_nominal, Id_Map_input_file, Num_components_Id_Map, Id_Map_q, Id_Map_t, Id_Map_e, input_data, Response_bit_length):	
	newdir_field = 'Field_results'			#Store Field Responses in this Director.
	os.makedirs(os.path.join(newdir, newdir_field))
	newdir_field = newdir + '/' + newdir_field
		
	Conditions_field = 0
	path = os.path.join(current_loc, field_directory)
	print('path =' + path)
	
	new_file = []	# Stores the file names in the field
	if not os.path.exists(field_directory):
		print('Field_directory not found\n')
	
	for information in os.walk(path):
		Conditions_field += len(information[2])	# Count the number of files in the field.
	
	for i in range(0,Conditions_field):
		if  (str(information[2][i]) != field_nominal):
			x=str.find(str(information[2][i]), '.csv')
			new_file.append( str(information[2][i])[:x])
		
			
	print('Total conditions at field ' + repr(Conditions_field-1))
	
	
	raw_data = []							# Store the data of field device.
	file_path = os.path.join(field_directory, field_nominal)
	file_read(file_path, raw_data)
	
	
	
	trim=str.find(field_nominal,'.csv')
	field_nominal= field_nominal[:trim]
	
	total_devices_field = len(raw_data)
	Field_response =[[] for i in range(total_devices_field)]
	response =[[] for i in range(total_devices_field)]
	
	print('total_devices ' + repr(total_devices_field))
	
	raw_data_f1 = [[] for i in range(Conditions_field-1)]	#Read all the files in the field, except the nominal one (e.g 1.2V_25C)
	for i in range(len(new_file)):	
		file_path = os.path.join(field_directory, (new_file[i]+'.csv'))
		file_read(file_path, raw_data_f1[i])
		
	
	
	for i in range(0,Conditions_field-1):
		for j in range(0,total_devices_field):
			identity_map(Id_Map_input_file,raw_data_f1[i][j], new_file[i]+'_Enrol_R_Id_Map.txt',new_file[i]+'_Field_R_Id_Map.txt', Field_response[j], j, Num_components_Id_Map, Id_Map_q, Id_Map_t, Id_Map_e, response[j], newdir_field, input_data)
		reliability(response, Field_response, new_file[i]+'_Reliability_id_map.txt', total_devices_field, Response_bit_length, newdir_field) #inputdata_nominal, inputdata_field, output_file, Num_devices_field, Response_lenght, directory_name
		
		Field_response =[[] for i in range(total_devices_field)]
		response =[[] for i in range(total_devices_field)]
	#sys.exit()
	return




if __name__ == "__main__":
	cProfile.run('main()', 'restats')